/*
 * Cool Technical Desire, general accounting and statistics API.
 *
 * Copyright (c) 2000, 2001 Lev Walkin, Alexandr Grigoryev, Vlad Skvortsov.
 *
 */
#ifndef	__DESIRE_H__
#define	__DESIRE_H__

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>

union CTB4 {
	/* i: To be accessed by whole (FL.i) */
	unsigned int	i;

	/* s: To be accessed by shorts (FL.s.s1 - first short) */
	struct _CTB_shorts {
		unsigned short s1	__attribute__ ((packed));
		unsigned short s2	__attribute__ ((packed));
	} s;

	/* c: To be accessed byte-wise
	** (FL.c.s1c1 - first byte of the first short)
	*/
	struct _CTB_bytes {
		unsigned char s1c1	__attribute__ ((packed));
		unsigned char s1c2	__attribute__ ((packed));
		unsigned char s2c1	__attribute__ ((packed));
		unsigned char s2c2	__attribute__ ((packed));
	} c;
};

/* Default data format */
typedef struct {
	time_t	TS	__attribute__ ((packed));	/* Time stamp */
	union CTB4 FL	__attribute__ ((packed));	/* Flags */
	unsigned int In __attribute__ ((packed));	/* In */
	unsigned int Out __attribute__ ((packed));	/* Out */
	unsigned int Src __attribute__ ((packed));	/* Source */
	unsigned int Dst __attribute__ ((packed));	/* Destination */
	unsigned int U1 __attribute__ ((packed));	/* User defined */
	unsigned int U2 __attribute__ ((packed));	/* User defined */
} ctb_data;

/* FIELD constants: field offsets within ctb_data structure */
#define	FIELD_TS	 0	/* TS Offset */
#define	FIELD_FL	 4	/* FL Offset */
#define	FIELD_IN	 8	/* In Offset */
#define	FIELD_OUT	12	/* Out Offset */
#define	FIELD_SRC	16	/* Src Offset */
#define	FIELD_DST	20	/* Dst Offset */
#define	FIELD_U1	24	/* U1 Offset */
#define	FIELD_U2	28	/* U2 Offset */

/* LP constants: accounting period position */
#define	CURR  0x00
#define	LAST  0x10
#define	PREV  0x20

/* HDWMY constants: accounting period */
#define	HOUR  0x01
#define	DAY   0x02
#define	WEEK  0x03
#define	MONTH 0x04
#define	YEAR  0x05

typedef struct {
	time_t from;
	time_t to;
	int simple_def;
		/*
		** either constants:
		** 0 - use from & to fields
		** 1 - use simple_def (LP | HDWMY)
		** 2 - both simple_def & from & to are set correctly
		*/
	int	either;
} ctb_range;

/******************************************/
/* Initialize and open accounting session */
/******************************************/
typedef void ctb_session;

ctb_session *
ctb_open(
	const char *optConfigName,	/* /usr/local/etc/desire.conf */
	const char *Service,	/* Service name */
	const char *optName	/* Optional subdestination name */
);

/* Change destination name (optName from ctb_open) within the service */
int	ctb_chname(ctb_session *, const char *name);
void	ctb_close(ctb_session *);

/* Get CTB session description */
char *ctb_description(ctb_session *);

/*
** Add this structure to the CTB file
** Note: it locks the last used file descriptor!
** Be sure to ctb_close() or ctb_chname() when you're finished!
*/
int	ctb_add(ctb_session *, void *data);

/* Get the specified data range of data */
void *ctb_get(ctb_session *, ctb_range *range, int *count);

/* Synchronize with on-disk state */
void ctb_sync(ctb_session *ctbs);

/********************/
/* Formatted output */
/********************/

char *ctb_format(ctb_session *, const char *format, void *data);

/*
 * Where *format is the string with the following metacharacters:
 *
 * %t	-	timestamp in unix time
 * %T	-	timestamp in human-readable form (ctime(3))
 * %f	-	flags in numeric form
 * %s	-	source field (Src) in numeric form.
 * %S	-	mapped source field (Src)
 * %d	-	destination field (Dst) in numeric form.
 * %D	-	mapped destination field (Dst)
 * %i	-	In
 * %o	-	Out
 * %I	-	In divided by 1024
 * %O	-	Out divided by 1024
 * %1	-	U1 field, mapped if mapping enabled
 * %2	-	U2 field, mapped if mapping enabled
 *
 */

/*****************/
/*** BYTE MAPS ***/
/*****************/

/*
** If string pointer is a NULL value, than map
** will be deleted instead of replacing.
** ctb_addmap() will return 0 in case of any error, or WandIndex.
*/
int
ctb_addmap(
	ctb_session *,
	int FIELD,
		/* Desired index to be assigned to string */
	unsigned int optWantIndex,
	const char *string
);

/* Map binary value to string */
char	*ctb_map(ctb_session *, void *data, int FIELD);

/* rvmodeFlag */
#define	RVM_EXACT	0x00
#define	RVM_FUZZY	0x01	/* Use fnmatch() on value */

/* Map string to binary value */
unsigned int ctb_revmap(
	ctb_session *,
	int FIELD,
	const char *value,
	int rvmodeFlag,
	int (*opt_fuzzy_return)(
		unsigned int Idx,
		char const *resolvedValue,
		void *optRecvKey
	),
	void *optSendKey
);


/********************/
/* Helper functions */
/********************/

/* Get entry chunk size for the specified session */
unsigned int ctb_entry_size(ctb_session *);

/* Get the last used number */
unsigned int ctb_map_lastno(ctb_session *, int FIELD);

/* Get the number of the entries in map */
ssize_t ctb_map_size(ctb_session *, int FIELD);


/*
** Fills specified time range with appropriate values
** according to LP_HDWMY.
*/
ctb_range *ctb_fill_time(ctb_range *optRange, int LP_HDWMY);


/*************************/
/*** Files multiplexor ***/
/*************************/

/*
** Read n files and multiplex their contents into one Desire session
** Note: Not all the contents of all specified files may be fully read.
** optPosExt specify the extensions to be used by position cache file.
*/
int ctb_mux(
	ctb_session *ctbs,
	ctb_data *(*ParserFunc)(char *StringToParse),
	int mux_flags,
	char *fnames[],
	char *optPosExt
);

/*
** Flags mux_flags:
*/
/* Ignore errors on file opens */
#define	CTB_MUX_IGN_FOF   0x01

/* Truncate file when EOF reached */
#define	CTB_MUX_TRUNC     0x02

/* Do not wait for other files that are ended (Dangerous) */
#define	CTB_MUX_NOWAIT    0x04


/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

/* Internal functions */

off_t	ctb_findstart (FILE *fp, time_t from, off_t fsize, int EntrySize);
off_t	ctb_findend   (FILE *fp, time_t till, off_t fsize, int EntrySize);

#endif

