#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "desire.h"
#include "ctbint.h"
#include "session.h"

typedef struct _filez {
	FILE *fd;
	int posfd;
	char const *fname;
	ctb_data *chunk;

	/* -1: EOF detected, 0: Data should be read, 1: ->chunk filled */
	int	actual;

	struct _filez *next;
} filez;

int
_ctb_close_filez(filez *f) {

	if(!f)
		return -1;

	/* Close all child filez */
	if(f->next)
		_ctb_close_filez(f->next);

	/* Closing main stream */
	if(f->fd)
		fclose(f->fd);

	/* Closing position cache file */
	if(f->posfd != -1)
		close(f->posfd);

	/* Freeing temporary ctb_data. */
	free(f->chunk);

	/* Any comments? */
	free(f);
	return 0;
}

int
_ctb_add2filez(ctb_sess *ctbs, filez **flz, char const *fname, char *optPosExt, int flags) {
	filez *f;
	struct stat sb;

	/* Allocating filez structur */
	f = (filez *)calloc(1, sizeof(filez));
	if(!f)	/* ENOMEM */
		return -1;

	/* Allocating temporary ctb_data of EntrySize */
	f->chunk = (ctb_data *)calloc(1, ctbs->EntrySize);
	if(!f->chunk) {
		free(f);
		return -1;
	}

	/*
	** Opening main log stream
	** If CTB_MUX_TRUNC, then open read-write
	** so we may later invoke ftruncate().
	*/
	f->fd = fopen(fname, (flags & CTB_MUX_TRUNC)?"r+":"r");
	if(!f->fd) {
		/* Can't open file */
		free(f->chunk);
		free(f);
		return -1;
	}

	/* Simple security checks */
	if( (fstat(fileno(f->fd), &sb) == -1)
		|| ((sb.st_mode & S_IFMT) != S_IFREG)
	) {
		fclose(f->fd);
		free(f->chunk);
		free(f);
		return -1;
	}

	/* Enabling position cache file */
	if( (f->posfd = _ctb_initFILEpos(fname, f->fd, optPosExt)) == -1 ) {
		fclose(f->fd);
		free(f->chunk);
		free(f);
		return -1;
	}

	/* Add to the head */
	f->next = *flz;	/* NULL, if not initialized */
	*flz = f;

	f->fname = fname;

	return 0;
}

/*
** _ctb_mux_getnextvalid()
**
** This function intended to fill the f->chunk data with the next string
** from the associated stream f->fd.
**
** f->actual specifies whether data in f->chunk made actual,
** or end of file reached.
**
*/

int
_ctb_mux_getnextvalid(_ctb_sbuf *sb, filez *f,
		ctb_data *(*func)(char *), int flags, int es) {
	char *p;
	ctb_data *data;

	while( !_ctb_sbuf_zero(sb) && (p = _ctb_sbuf_fgets(sb, f->fd)) ) {
		if(!(data = func(p)))
			continue;

		memcpy(f->chunk, data, es);

		f->actual = 1;

		return 0;
	}

	/*
	** Ftruncate if file is fully read
	** Bugs:
	**   With this simple implementation of truncating the file if
	**   it is fully read, only one file by one invocation of ctb_mux
	**   can be truncated. If >1 files ends with the same time stamp,
	**   only one of them will be truncated.
	*/
	if((flags & CTB_MUX_TRUNC)) {
		/* f->fd should be opened "r+" at this moment */
		ftruncate(fileno(f->fd), 0);
		fseek(f->fd, 0, SEEK_SET);
		_ctb_adjFILEpos(f->fd, f->posfd, 0);
	}

	f->actual = -1;

	return -1;
}

int
_ctb_mux_loop(ctb_sess *ctbs, filez **files, ctb_data *(*func)(char *), int flags) {
	filez *f;
	time_t mintime;
	filez *minfile;
	_ctb_sbuf *sb;
	int es = ctbs->EntrySize;

	sb = _ctb_sbuf_init();

  while(1) {

	mintime = 0;
	minfile = NULL;

	for(f = *files; f; f = f->next) {

		/* Ignore empty streams */
		if(f->actual == -1)
			continue;

		/* Add some bytes to empty structure */
		if(!f->actual) {
			if( _ctb_mux_getnextvalid(sb, f, func, flags, es) == -1 )
				return -1;
		}

		if(!minfile || mintime > f->chunk->TS) {
			mintime = f->chunk->TS;
			minfile = f;
		}
	}

	/* No available files found */
	if(!minfile) {
		return -1;
	}

	/*
	** Add the information from the specified stream to desire.
	** Adjust that stream's position pointer.
	*/
	ctb_add(ctbs, minfile->chunk);
	_ctb_adjFILEpos(minfile->fd, minfile->posfd, 0);

	if(minfile->actual == 1)
		minfile->actual = 0;

 };

	_ctb_sbuf_free(sb);

	return 0;
}

int
ctb_mux(ctb_session *ctbsp, ctb_data *(*func)(char *), int flags, char *fnames[], char *optPosExt) {
	register ctb_sess *ctbs = (ctb_sess *)ctbsp;
	filez *files = NULL;

	/* Common security checks */
	if(!ctbsp || !func || !fnames) {
		errno = EINVAL;
		return -1;
	}

	/* Position cache file extension checks */
	if(!optPosExt) {
		optPosExt = "desire-mux.pos";
	} else {
		if(!*optPosExt || strchr(optPosExt, '/')) {
			errno = EINVAL;
			return -1;
		}
	}

	/* Open specified files and related position cache files */
	for(; *fnames; fnames++) {
		if(_ctb_add2filez(ctbs, &files, *fnames, optPosExt, flags)
			&& !(flags & CTB_MUX_IGN_FOF)) {
			_ctb_close_filez(files);
			return -1;
		}
	}

	/* Successfully processed 0 files %-O */
	if(!files)
		return 0;

	/* Enter main processing loop */
	_ctb_mux_loop(ctbs, &files, func, flags);

	/* Close the rest */
	_ctb_close_filez(files);

	return 0;
}


